"""
GIT tag-based version information manager
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import re
from pykrete.args import environ
from pykrete.repo import Git
from .version import Version
from .ci_io import CiIo
from .ci_revision import CiRevision
from .bumper import Bumper


class TagVersion(Version):
    """Handle version in CI environment
    A version is constructed from the last version tag (tag name format '<major>.<minor>.<build>'
     indicating a release version, with a message format of 'ci_base_build <build>' indicating the
     CI build number for the first release with this major-minor combo).
    Build is calculated as the current CI project-build# variable, minus the build specified in the
     last version tag's message (unless bumper was activated, in which case it'll be 0).
    Revision is read from the CI environment variables (see CiRevision).
    """
    def __init__(self, tag_reader=None, bumper=Bumper.disabled, ci_reader=environ, ci_spec=None):
        """Initializes this instance from CI environment

        :param tag_reader: (pattern, is_must)=>tag list
         [see pykrete.repo.git.get_tags_from behavior]
        :param bumper: (version, tags)=>version, was_bumped
         [see pykrete.versioning.Bumper.major_minor_tag behavior]
        :param ci_reader: (ci_variable, role)=>value function
         [see pykrete.args.environ behavior]
        :param ci_spec: dictionary of part to CI variable
         (parts are: 'major version', 'minor version', 'build version', 'branch name',
          'merge request title')
        """
        self.__ci_io = CiIo(ci_reader, ci_spec)
        self.__tag_reader = tag_reader if tag_reader else Git().get_tags_from
        self.__bumper = bumper
        revision_type = CiRevision(self.__ci_io).revision_type
        (major, minor, build) = self._read_version_from_tags()
        super().__init__(major=major,
                         minor=minor,
                         revision=revision_type.value,
                         build=build,
                         revision_type=revision_type)

    def _read_version_from_tags(self):
        tags = self.__tag_reader(pattern=r'(\d+\.){2}\d+', is_must=False)
        last_version_tag = tags[0] if tags else None
        version, base_build = self.__get_version_and_base_build_from(last_version_tag)
        version, was_bumped = self.__bumper(version, tags)
        if not was_bumped:
            version = self.__bump_relative_build(version, base_build)
        return tuple(version)

    def __bump_relative_build(self, version, base_build):
        ci_build = int(self.__ci_io.read_env('build version'))
        version[2] = ci_build - base_build
        return version

    @staticmethod
    def __get_version_and_base_build_from(tag):
        if not tag:
            return [0, 0, 0], 0
        version = [int(part) for part in tag[0].split('.')]
        message_build = re.findall(r'ci_base_build (\d+)', tag[1]) if tag[1] else None
        base_build = int(message_build[0]) if message_build else 0
        return version, base_build



"""
Version information managers
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from .ci_version import CiVersion
from .revision_type import RevisionType
from .version import Version
from .version_py_version import VersionPyVersion
from .tag_version import TagVersion
from .bumper import Bumper

__all__ = ['CiVersion', 'RevisionType', 'Version', 'VersionPyVersion', 'TagVersion', 'Bumper']



"""
GIT tag-based version information manager
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""


class Bumper:
    """Bumpers static class"""

    @staticmethod
    def major_minor_tag(version, tags):
        """Locates 'bump' tags (tag name is either 'bump_minor' or 'bump_major') then increments the
         relevant version part.
        One bump at most is performed, with bump_major overriding bump_minor.
        :param version: original version
        :param tags: tags read after original version
        :return: (version, was_bumped) the calculated versions, and a bool flag indicating if
         bumping occurred
        """
        bumps = [tag[0].split('_')[1] for tag in tags if tag and tag[0].startswith('bump_')]
        if 'major' in bumps:
            return [version[0] + 1, 0, 0], True
        if 'minor' in bumps:
            return [version[0], version[1] + 1, 0], True
        return version, False

    @staticmethod
    def disabled(version, tags):    # pylint: disable=unused-argument
        """Keep version as-is regardless of tags"""
        return version, False



"""
Pykrete versioning.tag_version tests
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

import unittest
import logging
from pykrete.versioning import TagVersion, RevisionType, Version, Bumper
from .versioning import PykreteVersioningTestCase


class PykreteVersioningTagVersionTestCase(PykreteVersioningTestCase):
    """Unit tests for pykrete's versioning module's TagVersion class
    """

    _logger = logging.getLogger(__name__)
    __tags = []

    def test_tag_version_server(self):
        """Verifies handling of CI-environment version
        NOTE: THIS TEST WILL FAIL LOCALLY, SAFE TO IGNORE
        """
        self._logger.info('NOTE: THIS TEST WILL FAIL LOCALLY, SAFE TO IGNORE')
        self._assert_version_pattern(TagVersion())

    def test_tag_version_local_alpha_no_tags(self):
        """Verifies handling of no simulated tags present"""
        self._assert_tag_version_local(
            spec={'build version': '3',
                  'branch name': 'custom',
                  'merge request title': 'None'}, tags=[None],
            expected_version=Version((0, 0, 1, 3, RevisionType.Alpha)))

    def test_tag_version_local_alpha_first_minor_bump(self):
        """Verifies handling of simulated minor bump without a previous release"""
        self._assert_tag_version_local(
            spec={'build version': '4',
                  'branch name': 'custom',
                  'merge request title': 'None'},
            tags=[None, ('bump_minor', None)],
            expected_version=Version((0, 1, 1, 0, RevisionType.Alpha)))

    def test_tag_version_local_beta_major_bump(self):
        """Verifies handling of simulated major bump after a previous release"""
        self._assert_tag_version_local(
            spec={'build version': '111',
                  'branch name': 'custom',
                  'merge request title': 'WIP: something'},
            tags=[('0.7.56', 'ci_base_build 106'), ('bump_major', None)],
            expected_version=Version((1, 0, 2, 0, RevisionType.Beta)))

    def test_tag_version_local_rc_disabled_bump(self):
        """Verifies handling of simulated minor after major bump after a previous release
        where only the major bump should be used
        """
        self._assert_tag_version_local(
            spec={'build version': '1234',
                  'branch name': 'custom',
                  'merge request title': 'something'},
            tags=[('1.2.34', 'ci_base_build 1198'), ('bump_major', None), ('bump_minor', None)],
            expected_version=Version((1, 2, 3, 36, RevisionType.RC)),
            bump=False)

    def test_tag_version_local_rc_two_minor_bumps(self):
        """Verifies handling of simulated two minor bumps after a previous release
        where only one minor bump should be used
        """
        self._assert_tag_version_local(
            spec={'build version': '12345',
                  'branch name': 'custom',
                  'merge request title': 'something'},
            tags=[('11.22.341', 'ci_base_build 12338'), ('bump_minor', None), ('bump_minor', None)],
            expected_version=Version((11, 23, 3, 0, RevisionType.RC)))

    def test_tag_version_local_release_two_major_bumps(self):
        """Verifies handling of simulated two major bumps after a previous release
        where only one major bump should be used
        """
        self._assert_tag_version_local(
            spec={'build version': '88888',
                  'branch name': 'master',
                  'merge request title': 'blaster'},
            tags=[('7.7.77', 'ci_base_build 88880'), ('bump_major', None), ('bump_major', None)],
            expected_version=Version((8, 0, 4, 0, RevisionType.Release)))

    def _assert_tag_version_local(self, spec, tags, expected_version, bump=True):
        self.__tags = tags
        target = TagVersion(self._reader, Bumper.major_minor_tag if bump else Bumper.disabled,
                            self._echo, spec)
        self._assert_version_pattern(target)
        self.assertEqual(expected_version, target,
                         f'version not as expected: should be {expected_version}')

    def _reader(self, pattern, is_must):
        if is_must:
            raise AssertionError('tags are fetched with is_must=True')
        if not pattern:
            raise AssertionError('no pattern was provider for fetching tags')
        return self.__tags


if __name__ == '__main__':
    unittest.main()
